"""
Tests for the random module.
"""
import pytest
from context import stag
import stag.random


def test_sbm():
    # Generate a graph with a fixed p and q
    n = 2000
    p = 0.8
    q = 0.2
    graph = stag.random.sbm(n, 4, p, q, exact=True)
    assert graph.number_of_vertices() == 2000

    # The adjacency matrix should be symmetric
    sym_diff = (graph.adjacency() - graph.adjacency().transpose())
    sym_diff.eliminate_zeros()
    assert sym_diff.nnz == 0

    # The number of edges should be about
    # n * ((n/4) * p + (3 * n/4) * q)
    assert abs((graph.total_volume() / (n * ((n/4) * p + (3 * n/4) * q))) - 1) <= 0.1

    # The approximate version should give approximately the same number of edges
    graphapx = stag.random.sbm(n, 4, p, q)
    assert graphapx.number_of_vertices() == n
    assert abs((graphapx.number_of_edges() / graph.number_of_edges()) - 1) <= 0.2


def test_erdos_renyi():
    # Generate a graph
    n = 1000
    graph = stag.random.erdos_renyi(n, 0.1, exact=True)

    # Check that the graph has the expected number of vertices and edges.
    assert graph.number_of_vertices() == n
    assert abs((graph.total_volume() / (int(2 * 0.1 * (n * (n - 1)) / 2) + n)) - 1) <= 0.1



from . import stag_internal
from . import graph
from .utility import return_graph


@return_graph
def load_edgelist(filename: str):
    """
    Load a graph from an edgelist file.

    We define an edgelist file in the following way.
      - Any lines beginning with '#' or '//' are ignored
      - Any blank lines are ignored
      - All other lines are of some format:
          <u>, <v>, <weight>
          <u>, <v>
          <u> <v> <weight>
          <u> <v>
        where <u> and <v> can be parsed as integers, and <weight> can be parsed
        as a double. If the weight is ommitted, it is taken to be 1. The
        data lines of the edgelist file should all have the same format.

    :return: a STAG graph object constructed from the given edgelist file
    :except: error if the file cannot be parsed as an edgelist
    """
    return stag_internal.load_edgelist(filename)


def save_edgelist(g: graph.Graph, filename: str):
    """
    Save the given graph as an edgelist file.

    :param g: the graph object to be saved
    :param filename: the name of the file to save the edgelist data to
    """
    stag_internal.save_edgelist(g.internal_graph, filename)



# This file was automatically generated by SWIG (http://www.swig.org).
# Version 4.0.2
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.

from sys import version_info as _swig_python_version_info
if _swig_python_version_info < (2, 7, 0):
    raise RuntimeError("Python 2.7 or later required")

# Import the low-level C/C++ module
if __package__ or "." in __name__:
    from . import _stag_internal
else:
    import _stag_internal

try:
    import builtins as __builtin__
except ImportError:
    import __builtin__

def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except __builtin__.Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)


def _swig_setattr_nondynamic_instance_variable(set):
    def set_instance_attr(self, name, value):
        if name == "thisown":
            self.this.own(value)
        elif name == "this":
            set(self, name, value)
        elif hasattr(self, name) and isinstance(getattr(type(self), name), property):
            set(self, name, value)
        else:
            raise AttributeError("You cannot add instance attributes to %s" % self)
    return set_instance_attr


def _swig_setattr_nondynamic_class_variable(set):
    def set_class_attr(cls, name, value):
        if hasattr(cls, name) and not isinstance(getattr(cls, name), property):
            set(cls, name, value)
        else:
            raise AttributeError("You cannot add class attributes to %s" % cls)
    return set_class_attr


def _swig_add_metaclass(metaclass):
    """Class decorator for adding a metaclass to a SWIG wrapped class - a slimmed down version of six.add_metaclass"""
    def wrapper(cls):
        return metaclass(cls.__name__, cls.__bases__, cls.__dict__.copy())
    return wrapper


class _SwigNonDynamicMeta(type):
    """Meta class to enforce nondynamic attributes (no new attributes) for a class"""
    __setattr__ = _swig_setattr_nondynamic_class_variable(type.__setattr__)


import weakref

class SwigPyIterator(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined - class is abstract")
    __repr__ = _swig_repr
    __swig_destroy__ = _stag_internal.delete_SwigPyIterator

    def value(self):
        return _stag_internal.SwigPyIterator_value(self)

    def incr(self, n=1):
        return _stag_internal.SwigPyIterator_incr(self, n)

    def decr(self, n=1):
        return _stag_internal.SwigPyIterator_decr(self, n)

    def distance(self, x):
        return _stag_internal.SwigPyIterator_distance(self, x)

    def equal(self, x):
        return _stag_internal.SwigPyIterator_equal(self, x)

    def copy(self):
        return _stag_internal.SwigPyIterator_copy(self)

    def next(self):
        return _stag_internal.SwigPyIterator_next(self)

    def __next__(self):
        return _stag_internal.SwigPyIterator___next__(self)

    def previous(self):
        return _stag_internal.SwigPyIterator_previous(self)

    def advance(self, n):
        return _stag_internal.SwigPyIterator_advance(self, n)

    def __eq__(self, x):
        return _stag_internal.SwigPyIterator___eq__(self, x)

    def __ne__(self, x):
        return _stag_internal.SwigPyIterator___ne__(self, x)

    def __iadd__(self, n):
        return _stag_internal.SwigPyIterator___iadd__(self, n)

    def __isub__(self, n):
        return _stag_internal.SwigPyIterator___isub__(self, n)

    def __add__(self, n):
        return _stag_internal.SwigPyIterator___add__(self, n)

    def __sub__(self, *args):
        return _stag_internal.SwigPyIterator___sub__(self, *args)
    def __iter__(self):
        return self

# Register SwigPyIterator in _stag_internal:
_stag_internal.SwigPyIterator_swigregister(SwigPyIterator)

class vectori(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def iterator(self):
        return _stag_internal.vectori_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self):
        return _stag_internal.vectori___nonzero__(self)

    def __bool__(self):
        return _stag_internal.vectori___bool__(self)

    def __len__(self):
        return _stag_internal.vectori___len__(self)

    def __getslice__(self, i, j):
        return _stag_internal.vectori___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _stag_internal.vectori___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _stag_internal.vectori___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _stag_internal.vectori___delitem__(self, *args)

    def __getitem__(self, *args):
        return _stag_internal.vectori___getitem__(self, *args)

    def __setitem__(self, *args):
        return _stag_internal.vectori___setitem__(self, *args)

    def pop(self):
        return _stag_internal.vectori_pop(self)

    def append(self, x):
        return _stag_internal.vectori_append(self, x)

    def empty(self):
        return _stag_internal.vectori_empty(self)

    def size(self):
        return _stag_internal.vectori_size(self)

    def swap(self, v):
        return _stag_internal.vectori_swap(self, v)

    def begin(self):
        return _stag_internal.vectori_begin(self)

    def end(self):
        return _stag_internal.vectori_end(self)

    def rbegin(self):
        return _stag_internal.vectori_rbegin(self)

    def rend(self):
        return _stag_internal.vectori_rend(self)

    def clear(self):
        return _stag_internal.vectori_clear(self)

    def get_allocator(self):
        return _stag_internal.vectori_get_allocator(self)

    def pop_back(self):
        return _stag_internal.vectori_pop_back(self)

    def erase(self, *args):
        return _stag_internal.vectori_erase(self, *args)

    def __init__(self, *args):
        _stag_internal.vectori_swiginit(self, _stag_internal.new_vectori(*args))

    def push_back(self, x):
        return _stag_internal.vectori_push_back(self, x)

    def front(self):
        return _stag_internal.vectori_front(self)

    def back(self):
        return _stag_internal.vectori_back(self)

    def assign(self, n, x):
        return _stag_internal.vectori_assign(self, n, x)

    def resize(self, *args):
        return _stag_internal.vectori_resize(self, *args)

    def insert(self, *args):
        return _stag_internal.vectori_insert(self, *args)

    def reserve(self, n):
        return _stag_internal.vectori_reserve(self, n)

    def capacity(self):
        return _stag_internal.vectori_capacity(self)
    __swig_destroy__ = _stag_internal.delete_vectori

# Register vectori in _stag_internal:
_stag_internal.vectori_swigregister(vectori)

class vectorl(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def iterator(self):
        return _stag_internal.vectorl_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self):
        return _stag_internal.vectorl___nonzero__(self)

    def __bool__(self):
        return _stag_internal.vectorl___bool__(self)

    def __len__(self):
        return _stag_internal.vectorl___len__(self)

    def __getslice__(self, i, j):
        return _stag_internal.vectorl___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _stag_internal.vectorl___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _stag_internal.vectorl___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _stag_internal.vectorl___delitem__(self, *args)

    def __getitem__(self, *args):
        return _stag_internal.vectorl___getitem__(self, *args)

    def __setitem__(self, *args):
        return _stag_internal.vectorl___setitem__(self, *args)

    def pop(self):
        return _stag_internal.vectorl_pop(self)

    def append(self, x):
        return _stag_internal.vectorl_append(self, x)

    def empty(self):
        return _stag_internal.vectorl_empty(self)

    def size(self):
        return _stag_internal.vectorl_size(self)

    def swap(self, v):
        return _stag_internal.vectorl_swap(self, v)

    def begin(self):
        return _stag_internal.vectorl_begin(self)

    def end(self):
        return _stag_internal.vectorl_end(self)

    def rbegin(self):
        return _stag_internal.vectorl_rbegin(self)

    def rend(self):
        return _stag_internal.vectorl_rend(self)

    def clear(self):
        return _stag_internal.vectorl_clear(self)

    def get_allocator(self):
        return _stag_internal.vectorl_get_allocator(self)

    def pop_back(self):
        return _stag_internal.vectorl_pop_back(self)

    def erase(self, *args):
        return _stag_internal.vectorl_erase(self, *args)

    def __init__(self, *args):
        _stag_internal.vectorl_swiginit(self, _stag_internal.new_vectorl(*args))

    def push_back(self, x):
        return _stag_internal.vectorl_push_back(self, x)

    def front(self):
        return _stag_internal.vectorl_front(self)

    def back(self):
        return _stag_internal.vectorl_back(self)

    def assign(self, n, x):
        return _stag_internal.vectorl_assign(self, n, x)

    def resize(self, *args):
        return _stag_internal.vectorl_resize(self, *args)

    def insert(self, *args):
        return _stag_internal.vectorl_insert(self, *args)

    def reserve(self, n):
        return _stag_internal.vectorl_reserve(self, n)

    def capacity(self):
        return _stag_internal.vectorl_capacity(self)
    __swig_destroy__ = _stag_internal.delete_vectorl

# Register vectorl in _stag_internal:
_stag_internal.vectorl_swigregister(vectorl)

class vectord(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def iterator(self):
        return _stag_internal.vectord_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self):
        return _stag_internal.vectord___nonzero__(self)

    def __bool__(self):
        return _stag_internal.vectord___bool__(self)

    def __len__(self):
        return _stag_internal.vectord___len__(self)

    def __getslice__(self, i, j):
        return _stag_internal.vectord___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _stag_internal.vectord___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _stag_internal.vectord___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _stag_internal.vectord___delitem__(self, *args)

    def __getitem__(self, *args):
        return _stag_internal.vectord___getitem__(self, *args)

    def __setitem__(self, *args):
        return _stag_internal.vectord___setitem__(self, *args)

    def pop(self):
        return _stag_internal.vectord_pop(self)

    def append(self, x):
        return _stag_internal.vectord_append(self, x)

    def empty(self):
        return _stag_internal.vectord_empty(self)

    def size(self):
        return _stag_internal.vectord_size(self)

    def swap(self, v):
        return _stag_internal.vectord_swap(self, v)

    def begin(self):
        return _stag_internal.vectord_begin(self)

    def end(self):
        return _stag_internal.vectord_end(self)

    def rbegin(self):
        return _stag_internal.vectord_rbegin(self)

    def rend(self):
        return _stag_internal.vectord_rend(self)

    def clear(self):
        return _stag_internal.vectord_clear(self)

    def get_allocator(self):
        return _stag_internal.vectord_get_allocator(self)

    def pop_back(self):
        return _stag_internal.vectord_pop_back(self)

    def erase(self, *args):
        return _stag_internal.vectord_erase(self, *args)

    def __init__(self, *args):
        _stag_internal.vectord_swiginit(self, _stag_internal.new_vectord(*args))

    def push_back(self, x):
        return _stag_internal.vectord_push_back(self, x)

    def front(self):
        return _stag_internal.vectord_front(self)

    def back(self):
        return _stag_internal.vectord_back(self)

    def assign(self, n, x):
        return _stag_internal.vectord_assign(self, n, x)

    def resize(self, *args):
        return _stag_internal.vectord_resize(self, *args)

    def insert(self, *args):
        return _stag_internal.vectord_insert(self, *args)

    def reserve(self, n):
        return _stag_internal.vectord_reserve(self, n)

    def capacity(self):
        return _stag_internal.vectord_capacity(self)
    __swig_destroy__ = _stag_internal.delete_vectord

# Register vectord in _stag_internal:
_stag_internal.vectord_swigregister(vectord)

class TupleMM(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self, arg2, arg3):
        _stag_internal.TupleMM_swiginit(self, _stag_internal.new_TupleMM(arg2, arg3))

    def get0(self):
        return _stag_internal.TupleMM_get0(self)

    def get1(self):
        return _stag_internal.TupleMM_get1(self)

    def set0(self, val):
        return _stag_internal.TupleMM_set0(self, val)

    def set1(self, val):
        return _stag_internal.TupleMM_set1(self, val)

    def __len__(self):
        return _stag_internal.TupleMM___len__(self)

    #[7]
    def __getitem__(self, n):
        if n >= len(self): raise IndexError()
        return getattr(self, 'get%d' % n)()
    def __setitem__(self, n, val):
        if n >= len(self): raise IndexError()
        getattr(self, 'set%d' % n)(val)

    __swig_destroy__ = _stag_internal.delete_TupleMM

# Register TupleMM in _stag_internal:
_stag_internal.TupleMM_swigregister(TupleMM)

class edge(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    v1 = property(_stag_internal.edge_v1_get, _stag_internal.edge_v1_set)
    v2 = property(_stag_internal.edge_v2_get, _stag_internal.edge_v2_set)
    weight = property(_stag_internal.edge_weight_get, _stag_internal.edge_weight_set)

    def __init__(self):
        _stag_internal.edge_swiginit(self, _stag_internal.new_edge())
    __swig_destroy__ = _stag_internal.delete_edge

# Register edge in _stag_internal:
_stag_internal.edge_swigregister(edge)
cvar = _stag_internal.cvar
VERSION_MAJOR = cvar.VERSION_MAJOR
VERSION_MINOR = cvar.VERSION_MINOR
VERSION_PATCH = cvar.VERSION_PATCH

class LocalGraph(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined - class is abstract")
    __repr__ = _swig_repr

    def degree(self, v):
        return _stag_internal.LocalGraph_degree(self, v)

    def degree_unweighted(self, v):
        return _stag_internal.LocalGraph_degree_unweighted(self, v)

    def neighbors(self, v):
        return _stag_internal.LocalGraph_neighbors(self, v)

    def neighbors_unweighted(self, v):
        return _stag_internal.LocalGraph_neighbors_unweighted(self, v)
    __swig_destroy__ = _stag_internal.delete_LocalGraph

# Register LocalGraph in _stag_internal:
_stag_internal.LocalGraph_swigregister(LocalGraph)

class Graph(LocalGraph):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self, *args):
        _stag_internal.Graph_swiginit(self, _stag_internal.new_Graph(*args))

    def adjacency(self):
        return _stag_internal.Graph_adjacency(self)

    def laplacian(self):
        return _stag_internal.Graph_laplacian(self)

    def normalised_laplacian(self):
        return _stag_internal.Graph_normalised_laplacian(self)

    def degree_matrix(self):
        return _stag_internal.Graph_degree_matrix(self)

    def inverse_degree_matrix(self):
        return _stag_internal.Graph_inverse_degree_matrix(self)

    def lazy_random_walk_matrix(self):
        return _stag_internal.Graph_lazy_random_walk_matrix(self)

    def total_volume(self):
        return _stag_internal.Graph_total_volume(self)

    def number_of_vertices(self):
        return _stag_internal.Graph_number_of_vertices(self)

    def number_of_edges(self):
        return _stag_internal.Graph_number_of_edges(self)

    def degree(self, v):
        return _stag_internal.Graph_degree(self, v)

    def degree_unweighted(self, v):
        return _stag_internal.Graph_degree_unweighted(self, v)

    def neighbors(self, v):
        return _stag_internal.Graph_neighbors(self, v)

    def neighbors_unweighted(self, v):
        return _stag_internal.Graph_neighbors_unweighted(self, v)
    __swig_destroy__ = _stag_internal.delete_Graph

# Register Graph in _stag_internal:
_stag_internal.Graph_swigregister(Graph)


def __eq__(*args):
    return _stag_internal.__eq__(*args)

def __ne__(*args):
    return _stag_internal.__ne__(*args)

def cycle_graph(n):
    return _stag_internal.cycle_graph(n)

def complete_graph(n):
    return _stag_internal.complete_graph(n)

def barbell_graph(n):
    return _stag_internal.barbell_graph(n)

def sprsMatValues(matrix):
    return _stag_internal.sprsMatValues(matrix)

def sprsMatInnerIndices(matrix):
    return _stag_internal.sprsMatInnerIndices(matrix)

def sprsMatOuterStarts(matrix):
    return _stag_internal.sprsMatOuterStarts(matrix)

def sprsMatToVec(*args):
    return _stag_internal.sprsMatToVec(*args)

def sprsMatFromVectors(column_starts, row_indices, values):
    return _stag_internal.sprsMatFromVectors(column_starts, row_indices, values)

def isSymmetric(matrix):
    return _stag_internal.isSymmetric(matrix)

def local_cluster(graph, seed_vertex, target_volume):
    return _stag_internal.local_cluster(graph, seed_vertex, target_volume)

def local_cluster_acl(*args):
    return _stag_internal.local_cluster_acl(*args)

def approximate_pagerank(graph, seed_vector, alpha, epsilon):
    return _stag_internal.approximate_pagerank(graph, seed_vector, alpha, epsilon)

def sweep_set_conductance(graph, vec):
    return _stag_internal.sweep_set_conductance(graph, vec)

def load_edgelist(filename):
    return _stag_internal.load_edgelist(filename)

def save_edgelist(graph, filename):
    return _stag_internal.save_edgelist(graph, filename)

def sbm(*args):
    return _stag_internal.sbm(*args)

def erdos_renyi(*args):
    return _stag_internal.erdos_renyi(*args)
class SprsMat(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    __swig_destroy__ = _stag_internal.delete_SprsMat

    def __init__(self):
        _stag_internal.SprsMat_swiginit(self, _stag_internal.new_SprsMat())

# Register SprsMat in _stag_internal:
_stag_internal.SprsMat_swigregister(SprsMat)

VERSION = _stag_internal.VERSION





from . import stag_internal
from . import utility


@utility.return_graph
def sbm(n: int, k: int, p: float, q: float, exact=False):
    """
    Generate a graph from the symmetric stochastic block model.

    Every cluster has the same number of vertices. For large enough values of
    n, this method samples from an approximate stochastic block model by
    default which significantly speeds up the execution time. To sample
    exactly from the stochastic block model, pass the optional 'exact'
    parameter to the method.

    The approximate sampling method has running time O(k^2 + nnz) and the exact
    method has running time O(n^2) where nnz is the number of non-zeros in the
    generated graph's adjacency matrix.

    :param n: - the number of vertices in the graph
    :param k: - the number of clusters; vertices are split evenly between clusters
    :param p: - the probability of including an edge inside a cluster
    :param q: - the probability of including an edge between two clusters
    :param exact: (optional) - whether to use the exact probability distribution
                              or an approximation.
    :return: the randomly generated graph
    """
    return stag_internal.sbm(n, k, p, q, exact)


@utility.return_graph
def erdos_renyi(n: int, p: float, exact=False):
    """
    Generate a graph from the Erdos-Renyi model.

    For large values of n, this method will use an approximate version of the
    random model with running time O(nnz) where nnz is the number of edges in
    the sampled graph.

    If the 'exact' parameter is true, then the true Erdos-Renyi distribution
    will be used, with running time O(n^2).

    :param n: - the number of vertices in the graph
    :param p: - the probability of including each edge
    :param exact: (optional) - whether to sample from the exact model
    :return: the randomly generated graph
    """
    return stag_internal.erdos_renyi(n, p, exact)



"""
Tests for handling reading and writing graphs to disk.
"""
import scipy as sp
import scipy.sparse
from context import stag
import stag.graph
import stag.random
import stag.graphio


def test_edgelist():
    ##########
    # TEST 1 #
    ##########
    # Let's load the different test graphs, and check that we get what we'd expect.
    expected_adj_mat = sp.sparse.csc_matrix([[0, 1, 1],
                                             [1, 0, 1],
                                             [1, 1, 0]])
    graph = stag.graphio.load_edgelist("data/test1.edgelist")
    adj_mat_diff = (graph.adjacency() - expected_adj_mat)
    adj_mat_diff.eliminate_zeros()
    assert adj_mat_diff.nnz == 0

    # Now save and reload the graph and check that the adjacency matrix has not changed
    stag.graphio.save_edgelist(graph, "data/temp.edgelist")
    graph = stag.graphio.load_edgelist("data/temp.edgelist")
    adj_mat_diff = (graph.adjacency() - expected_adj_mat)
    adj_mat_diff.eliminate_zeros()
    assert adj_mat_diff.nnz == 0

    ##########
    # TEST 2 #
    ##########
    expected_adj_mat = sp.sparse.csc_matrix([[0, 0.5, 0.5],
                                             [0.5, 0, 1],
                                             [0.5, 1, 0]])
    graph = stag.graphio.load_edgelist("data/test2.edgelist")
    adj_mat_diff = (graph.adjacency() - expected_adj_mat)
    adj_mat_diff.eliminate_zeros()
    assert adj_mat_diff.nnz == 0

    # Now save and reload the graph and check that the adjacency matrix has not changed
    stag.graphio.save_edgelist(graph, "data/temp.edgelist")
    graph = stag.graphio.load_edgelist("data/temp.edgelist")
    adj_mat_diff = (graph.adjacency() - expected_adj_mat)
    adj_mat_diff.eliminate_zeros()
    assert adj_mat_diff.nnz == 0

    ##########
    # TEST 3 #
    ##########
    expected_adj_mat = sp.sparse.csc_matrix([[0, 1, 0.5],
                                             [1, 0, 1],
                                             [0.5, 1, 0]])
    graph = stag.graphio.load_edgelist("data/test3.edgelist")
    adj_mat_diff = (graph.adjacency() - expected_adj_mat)
    adj_mat_diff.eliminate_zeros()
    assert adj_mat_diff.nnz == 0

    # Now save and reload the graph and check that the adjacency matrix has not changed
    stag.graphio.save_edgelist(graph, "data/temp.edgelist")
    graph = stag.graphio.load_edgelist("data/temp.edgelist")
    adj_mat_diff = (graph.adjacency() - expected_adj_mat)
    adj_mat_diff.eliminate_zeros()
    assert adj_mat_diff.nnz == 0

    ##########
    # TEST 4 #
    ##########
    expected_adj_mat = sp.sparse.csc_matrix([[0, 1, 0.5],
                                             [1, 0, 1],
                                             [0.5, 1, 0]])
    graph = stag.graphio.load_edgelist("data/test4.edgelist")
    adj_mat_diff = (graph.adjacency() - expected_adj_mat)
    adj_mat_diff.eliminate_zeros()
    assert adj_mat_diff.nnz == 0

    # Now save and reload the graph and check that the adjacency matrix has not changed
    stag.graphio.save_edgelist(graph, "data/temp.edgelist")
    graph = stag.graphio.load_edgelist("data/temp.edgelist")
    adj_mat_diff = (graph.adjacency() - expected_adj_mat)
    adj_mat_diff.eliminate_zeros()
    assert adj_mat_diff.nnz == 0



from . import stag_internal
import stag.graph
import scipy.sparse


def swig_sprs_to_scipy(swig_mat):
    """
    Take a swig sparse matrix and convert it to a scipy sparse matrix.
    """
    outer_starts = stag_internal.sprsMatOuterStarts(swig_mat)
    inner_indices = stag_internal.sprsMatInnerIndices(swig_mat)
    values = stag_internal.sprsMatValues(swig_mat)
    return scipy.sparse.csc_matrix((values, inner_indices, outer_starts))


def scipy_to_swig_sprs(scipy_mat: scipy.sparse.csc_matrix):
    """
    Take a scipy sparse matrix and convert it to a swig sprs matrix.
    """
    col_starts = stag_internal.vectorl(scipy_mat.indptr.tolist())
    row_indices = stag_internal.vectorl(scipy_mat.indices.tolist())
    values = stag_internal.vectord(scipy_mat.data.tolist())
    return stag_internal.sprsMatFromVectors(col_starts,
                                            row_indices,
                                            values)


def return_sparse_matrix(func):
    """
    A decorator which transforms a sparse matrix returned from the C++ library to a sparse scipy matrix for use within
    python. Note that this transformation incurs some overhead in terms of both time and space.

    :param func: the function whose output we would like to wrap
    :return: the decorated function
    """
    def decorated_function(*args, **kwargs):
        swig_sparse_matrix = func(*args, **kwargs)
        sp_sparse = swig_sprs_to_scipy(swig_sparse_matrix)
        del swig_sparse_matrix
        return sp_sparse

    return decorated_function


def return_graph(func):
    """
    A decorator which transforms a graph returned from the C++ library to the python version.

    :param func: the function whose output we would like to wrap
    :return: the decorated function
    """
    def decorated_function(*args, **kwargs):
        swig_graph = func(*args, **kwargs)
        return stag.graph.Graph(None, internal_graph=swig_graph)

    return decorated_function



import networkx
from abc import ABC, abstractmethod
from typing import List

from . import stag_internal
from . import utility


class Edge(object):
    """An object representing a weighted edge in a graph."""

    def __init__(self, v1, v2, weight):
        """
        :param v1: The first vertex in the edge.
        :param v2: The second vertex in the edge.
        :param weight: The weight of the edge.
        """
        self.v1 = v1
        self.v2 = v2
        self.weight = weight
        self.internal_edge = stag_internal.edge()
        self.internal_edge.v1 = v1
        self.internal_edge.v2 = v2
        self.internal_edge.weight = weight

    def __eq__(self, other):
        return self.v1 == other.v1 and self.v2 == other.v2 and self.weight == other.weight

    def __ne__(self, other):
        return not self.__eq__(other)


class LocalGraph(stag_internal.LocalGraph, ABC):
    """
    An abstract class representing local graph operations.
    For now, this class inherits directly from the internal local graph class.
    This 'intermediate' class is included for possible future python-specific functionality.
    """

    def __init__(self):
        pass

    @abstractmethod
    def degree(self, v) -> float:
        """Given a vertex v, return its weighted degree."""
        pass

    @abstractmethod
    def degree_unweighted(self, v) -> int:
        """
        Given a vertex v, return its unweighted degree. That is, the number of
        neighbors of v, ignoring the edge weights.
        """
        pass
    
    @abstractmethod
    def neighbors(self, v) -> List[Edge]:
        """
        Given a vertex v, return a vector of edges representing the
        neighborhood of v.

        The returned edges will all have the ordering (v, x) such that
        edge.v = v.

        :param v: an int representing some vertex in the graph
        :return: a list of Edge objects containing the neighborhood information
        """
        pass
    
    @abstractmethod
    def neighbors_unweighted(self, v) -> List[int]:
        """
        Given a vertex v, return a vector containing the neighbors of v.

        The weights of edges to the neighbors are not returned by this method.

        :param v: an int representing some vertex in the graph
        :return: an int vector giving the neighbors of v
        """
        pass


class Graph(LocalGraph):
    """
    Represents a graph. We keep things very simple - a graph is represented by its sparse adjacency matrix.

    In the general case, this allows for
      - directed and undirected graphs
      - self-loops

    If you'd like to store meta-data about the graph, such as node or edge labels, you should implement a subclass of
    this one, and add that information yourself.

    This graph cannot be dynamically updated. It must be initialised with the complete adjacency matrix.
    Vertices are referred to by their index in the adjacency matrix.
    """

    def __init__(self, adj_mat, internal_graph=None):
        """
        Initialise the graph with an adjacency matrix.

        :param adj_mat: A sparse scipy matrix.
        :param internal_graph: (optional) provide an internal STAG graph object with
                               which to initialise the python wrapper
        """
        # This class is essentially a thin wrapper around the stag_internal library, written in C++.
        if internal_graph is None:
            # Initialise the internal graph object with the provided adjacency matrix.
            adj_mat_csr = adj_mat.tocsr()
            outer_starts = stag_internal.vectorl(adj_mat_csr.indptr.tolist())
            inner_indices = stag_internal.vectorl(adj_mat_csr.indices.tolist())
            values = stag_internal.vectord(adj_mat_csr.data.tolist())
            self.internal_graph: stag_internal.Graph = stag_internal.Graph(outer_starts, inner_indices, values)
        else:
            # The initialiser was called with an internal graph object.
            self.internal_graph: stag_internal.Graph = internal_graph

        # Call the LocalGraph initialisation method.
        super().__init__()

    @utility.return_sparse_matrix
    def adjacency(self):
        """
        Return the sparse adjacency matrix of the graph.

        :return: a sparse matrix representing the graph adjacency matrix
        """
        return self.internal_graph.adjacency()

    @utility.return_sparse_matrix
    def laplacian(self):
        """
        Construct the Laplacian matrix of the graph.

        The Laplacian matrix is defined by
          L = D - A
        where D is the diagonal matrix of vertex degrees and A is the adjacency
        matrix of the graph.

        :return: a sparse matrix representing the graph Laplacian
        """
        return self.internal_graph.laplacian()

    @utility.return_sparse_matrix
    def normalised_laplacian(self):
        """
        Construct the normalised Laplacian matrix of the graph.

        The normalised Laplacian matrix is defined by
          Ln = D^{-1/2} L D^{-1/2}
        where D is the diagonal matrix of vertex degrees and L is the Laplacian
        matrix of the graph

        :return: a sparse matrix representing the normalised Laplacian
        """
        return self.internal_graph.normalised_laplacian()

    @utility.return_sparse_matrix
    def degree_matrix(self):
        """
        The degree matrix of the graph.

        The degree matrix is an n x n matrix such that each diagonal entry is the degree
        of the corresponding node.

        :return: the sparse degree matrix
        """
        return self.internal_graph.degree_matrix()
    
    @utility.return_sparse_matrix
    def inverse_degree_matrix(self):
        """
        The inverse degree matrix of the graph.

        The inverse degree matrix is an n x n matrix such that each diagonal entry is
        the inverse of the degree of the corresponding node, or 0 if the node
        has degree 0.

        :return: the sparse inverse degree matrix
        """
        return self.internal_graph.inverse_degree_matrix()
    
    @utility.return_sparse_matrix
    def lazy_random_walk_matrix(self):
        """
        The lazy random walk matrix of the graph.

        The lazy random walk matrix is defined to be
           1/2 I + 1/2 A D^{-1}
        where I is the identity matrix, A is the graph adjacency matrix and
        D is the degree matrix of the graph.

        :return: the sparse lazy random walk matrix
        """
        return self.internal_graph.lazy_random_walk_matrix()

    def total_volume(self):
        """
        The volume of the graph.

        The volume is defined as the sum of the node degrees.

        :return: the graph's volume.
        """
        return self.internal_graph.total_volume()

    def number_of_vertices(self):
        """The number of vertices in the graph."""
        return self.internal_graph.number_of_vertices()

    def number_of_edges(self):
        """
        The number of edges in the graph.

        This is defined based on the number of non-zero elements in the adjacency
        matrix, and ignores the weights of the edges.
        """
        return self.internal_graph.number_of_edges()

    def degree(self, vertex: int):
        return self.internal_graph.degree(vertex)

    def degree_unweighted(self, vertex: int):
        return self.internal_graph.degree_unweighted(vertex)

    def neighbors(self, vertex: int):
        return self.internal_graph.neighbors(vertex)

    def neighbors_unweighted(self, vertex: int):
        return self.internal_graph.neighbors_unweighted(vertex)

    def __eq__(self, other):
        #
        # IMPORTANT NOTE
        # Checking whether two graphs are the same is thought to be a 'difficult'
        # problem, and so this method does not really check for mathematical equivalence.
        # Rather, we just check that the adjacency matrices are the same. As such, this method
        # should not be relied on to test for graph isomorphism!
        #

        # Check basic size information about the two graphs first
        if self.number_of_vertices() != other.number_of_vertices():
            return False

        if self.number_of_edges() != other.number_of_edges():
            return False

        # Check that the data vectors of the graph adjacency matrices are equal.
        a1 = self.internal_graph.adjacency()
        a2 = other.internal_graph.adjacency()
        if stag_internal.sprsMatOuterStarts(a1) != stag_internal.sprsMatOuterStarts(a2):
            return False
        if stag_internal.sprsMatInnerIndices(a1) != stag_internal.sprsMatInnerIndices(a2):
            return False
        return stag_internal.sprsMatValues(a1) == stag_internal.sprsMatValues(a2)

    def __ne__(self, other):
        return not self.__eq__(other)

    def to_networkx(self):
        """
        Construct a networkx graph object which is equivalent to this STAG graph.
        """
        return networkx.Graph(self.adjacency())


@utility.return_graph
def cycle_graph(n):
    """
    Construct a cycle graph on n vertices.

    :param n:
    :return: a graph object representing the n-cycle
    """
    return stag_internal.cycle_graph(n)


@utility.return_graph
def complete_graph(n):
    """
    Construct a complete graph on n vertices.

    :param n:
    :return: a graph object representing the complete graph
    """
    return stag_internal.complete_graph(n)


@utility.return_graph
def barbell_graph(n):
    """
    Construct a barbell graph. The barbell graph consists of 2 cliques on n vertices,
    connected by a single edge.

    :param n:
    :return:
    """
    return stag_internal.barbell_graph(n)


def from_networkx(netx_graph, edge_weight_attribute="weight"):
    """
    Given a networkx graph, convert it to a STAG graph object.

    Unless otherwise specified, this method will use the 'weight' attribute on the
    networkx edges to assign the weight of the edges. If no such attribute is present,
    the edges will be added with weight 1.

    :param netx_graph: The networkx graph object to be converted.
    :param edge_weight_attribute: (default 'weight') the edge attribute to be used to
                                  generate the weights
    :return: A STAG graph object which is equivalent to the networkx graph.
    """
    return Graph(networkx.adjacency_matrix(netx_graph, weight=edge_weight_attribute))



"""Tests for the clustering algorithms."""
import scipy.sparse
import pytest
import numpy as np
from context import stag
import stag.graph
import stag.cluster

# Define the adjacency matrices of some useful graphs.
C4_ADJ_MAT = scipy.sparse.csc_matrix([[0, 1, 0, 1], [1, 0, 1, 0], [0, 1, 0, 1], [1, 0, 1, 0]])
K6_ADJ_MAT = scipy.sparse.csc_matrix([[0, 1, 1, 1, 1, 1], [1, 0, 1, 1, 1, 1], [1, 1, 0, 1, 1, 1],
                                      [1, 1, 1, 0, 1, 1], [1, 1, 1, 1, 0, 1], [1, 1, 1, 1, 1, 0]])
BARBELL5_ADJ_MAT = scipy.sparse.csc_matrix([[0, 1, 1, 1, 1, 0, 0, 0, 0, 0],
                                            [1, 0, 1, 1, 1, 0, 0, 0, 0, 0],
                                            [1, 1, 0, 1, 1, 0, 0, 0, 0, 0],
                                            [1, 1, 1, 0, 1, 0, 0, 0, 0, 0],
                                            [1, 1, 1, 1, 0, 1, 0, 0, 0, 0],
                                            [0, 0, 0, 0, 1, 0, 1, 1, 1, 1],
                                            [0, 0, 0, 0, 0, 1, 0, 1, 1, 1],
                                            [0, 0, 0, 0, 0, 1, 1, 0, 1, 1],
                                            [0, 0, 0, 0, 0, 1, 1, 1, 0, 1],
                                            [0, 0, 0, 0, 0, 1, 1, 1, 1, 0],
                                            ])


def test_default_local_clustering():
    # Construct a graph object with the barbell adjacency matrix
    graph = stag.graph.Graph(BARBELL5_ADJ_MAT)

    # Find a local cluster near the first vertex
    cluster = stag.cluster.local_cluster(graph, 1, 21)

    # Assert that the correct clusters have been found.
    assert (set(cluster) == {0, 1, 2, 3, 4})


def test_acl_local_clustering():
    # Construct a graph object with a well-defined cluster structure
    graph = stag.graph.barbell_graph(10)

    # Run the acl clustering method
    cluster = stag.cluster.local_cluster_acl(graph, 0, 0.9, 0.0001)

    # Check that we found one of the clusters
    expected_cluster = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}

    assert set(cluster) == expected_cluster


def test_approximate_pagerank():
    # For easier manual verification, we use a cycle graph with 0.5 weights on the edges
    adj = 0.5 * stag.graph.cycle_graph(4).adjacency()
    graph = stag.graph.Graph(adj)

    # Construct seed matrix.
    s = scipy.sparse.csc_matrix((4, 1))
    s[0, 0] = 1

    # Run the personalised pagerank and check that we get the right result
    p, r = stag.cluster.approximate_pagerank(graph, s, 1./3, 1./8)
    expected_p = [41./81, 2./27, 0, 2./27]
    expected_r = [5./81, 2./27 + 5./162, 2./27, 2./27 + 5./162]
    np.testing.assert_almost_equal(p.todense().transpose().tolist()[0], expected_p)
    np.testing.assert_almost_equal(r.todense().transpose().tolist()[0], expected_r)


def test_sweep_set():
    # Construct a simple graph to test with
    graph = stag.graph.barbell_graph(4)

    # Create the vector to test. The optimal conductance will be the first 4 vertices
    s = scipy.sparse.csc_matrix((8, 1))
    s[0, 0] = 0.1
    s[1, 0] = 0.25
    s[2, 0] = 0.2
    s[3, 0] = 0.15
    s[4, 0] = 0.05
    
    # Compute the sweep set
    sweep_set = stag.cluster.sweep_set_conductance(graph, s)
    assert set(sweep_set) == {0, 1, 2, 3}



import os
import platform

from setuptools import setup, find_packages
from distutils.core import setup, Extension

VERSION = '0.2.3'
DESCRIPTION = 'STAG: Spectral Toolkit of Algorithms for Graphs'
LONG_DESCRIPTION =\
    "This library provides several methods and algorithms relating to spectral graph theory in python."
URL = "https://staglibrary.io"

# Depending on the build platform, the required compiler flags are slightly
# different.
if platform.system() == 'Linux':
    compile_args = ['-std=c++2a']
elif platform.system() == 'Windows':
    compile_args = ['/std:c++20']
else:
    compile_args = ['-std=c++2a']

# specify the name of the extension and source files
# required to compile this
ext_modules = [Extension(name='stag._stag_internal',
                         sources=["stag/stag_internal_wrap.cxx",
                                  "stag/stag_lib/graph.cpp",
                                  "stag/stag_lib/random.cpp",
                                  "stag/stag_lib/graphio.cpp",
                                  "stag/stag_lib/cluster.cpp",
                                  "stag/stag_lib/utility.cpp"],
                         include_dirs=["stag/eigen-3.3.9", "stag/stag_lib"],
                         extra_compile_args=compile_args)]

# Setting up
setup(
    name="stag",
    ext_modules=ext_modules,
    version=VERSION,
    author="Peter Macgregor",
    author_email="<macgregor.pr@gmail.com>",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=["scipy", "networkx", "matplotlib"],
    long_description_content_type='text/markdown',
    url=URL,
    include_package_data=True,

    keywords=['python', 'spectral', 'graph', 'algorithms', 'clustering', 'cheeger'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Education",
        "Programming Language :: Python :: 3",
        # "Operating System :: MacOS :: MacOS X",
        # "Operating System :: Microsoft :: Windows",
        'Operating System :: POSIX :: Linux'
    ]
)



import scipy.sparse

from . import stag_internal
from . import graph
from . import utility


def local_cluster(g: graph.Graph, seed_vertex: int, target_volume: int):
    """
    Default local clustering algorithm. Given a graph and starting vertex,
    return a cluster which is close to the starting vertex.

    This method defaults to use the ACL local clustering algorithm.

    [ACL] Andersen, Reid, Fan Chung, and Kevin Lang.
    "Local graph partitioning using pagerank vectors." 2006
    47th Annual IEEE Symposium on Foundations of Computer Science (FOCS'06). IEEE, 2006.

    :param g: - a graph object implementing the LocalGraph interface
    :param seed_vertex: - the starting vertex in the graph
    :param target_volume: - the approximate volume of the cluster you would like to find
    :return: a vector containing the indices of vectors considered to be in the
            same cluster as the seed_vertex.
    """
    return stag_internal.local_cluster(g.internal_graph, seed_vertex, target_volume)


def local_cluster_acl(g: graph.Graph, seed_vertex: int, locality: float, error=0.001):
    """
    The ACL local clustering algorithm. Given a graph and starting vertex,
    returns a cluster close to the starting vertex, constructed in a local way.

    The locality parameter is passed as the alpha parameter in the personalised
    pagerank calculation.

    [ACL] Andersen, Reid, Fan Chung, and Kevin Lang.
    "Local graph partitioning using pagerank vectors." 2006
    47th Annual IEEE Symposium on Foundations of Computer Science (FOCS'06). IEEE, 2006.

    :param g: - a graph object implementing the LocalGraph interface
    :param seed_vertex: - the starting vertex in the graph
    :param locality: - a value in [0, 1] indicating how 'local' the cluster should
                      be. A value of '1' will return the return only the seed vertex
                      and a value of '0' will explore the whole graph.
    :param error: (optional) - the acceptable error in the calculation of the approximate
                              pagerank. Default 0.001.
    :return: a vector containing the indices of vectors considered to be in the
            same cluster as the seed_vertex.
    """
    return stag_internal.local_cluster_acl(g.internal_graph, seed_vertex, locality, error)


def approximate_pagerank(g: graph.Graph, seed_vector: scipy.sparse.csc_matrix, alpha: float, epsilon: float):
    """
    Compute the approximate pagerank vector as described in ACL:

    [ACL] Andersen, Reid, Fan Chung, and Kevin Lang.
    "Local graph partitioning using pagerank vectors." 2006
    47th Annual IEEE Symposium on Foundations of Computer Science (FOCS'06). IEEE, 2006.

    The parameters s, alpha, and epsilon are used as described in the paper.

    Note that the dimension of the returned vectors may not match the true
    number of vertices in the graph provided since the approximate
    pagerank is computed locally.

    :param g:
    :param seed_vector:
    :param alpha:
    :param epsilon:
    :return: A tuple of sparse column vectors corresponding to
               p - the approximate pagerank vector
               r - the residual vector
            By the definition of approximate pagerank, it is the case that
               p + pr(r, alpha) = pr(s, alpha)

    :raises ArgumentError: if the provided seed_vector is not a column vector.
    """
    apr = stag_internal.approximate_pagerank(g.internal_graph,
                                              utility.scipy_to_swig_sprs(seed_vector),
                                              alpha,
                                              epsilon)
    return utility.swig_sprs_to_scipy(apr[0]), utility.swig_sprs_to_scipy(apr[1])


def sweep_set_conductance(g: graph.Graph, vec: scipy.sparse.csc_matrix):
    """
    Find the sweep set of the given vector with the minimum conductance.
   
    First, sort the vector, and then let
        S_i = {v_j : j <= i}
    and return the set of original indices corresponding to
        argmin_i conducance(S_i)
   
    This method is expected to be run on vectors whose support is much less
    than the total size of the graph. If the total volume of the support of vec
    is larger than half of the volume of the total graph, then this method may
    return unexpected results.
   
    :param: graph
    :param: vec
    :return: a vector containing the indices of vec which give the minimum
            conductance in the given graph
    """
    return stag_internal.sweep_set_conductance(g.internal_graph, utility.scipy_to_swig_sprs(vec))



"""
Tests for the graph object.
"""
import numpy as np
import scipy as sp
import scipy.sparse
import networkx
import pytest
from context import stag
import stag.graph
import stag.random
import stag.graphio

# Define the adjacency matrices of some useful graphs.
C4_ADJ_MAT = scipy.sparse.csc_matrix([[0, 1, 0, 1], [1, 0, 1, 0], [0, 1, 0, 1], [1, 0, 1, 0]])
K6_ADJ_MAT = scipy.sparse.csc_matrix([[0, 1, 1, 1, 1, 1], [1, 0, 1, 1, 1, 1], [1, 1, 0, 1, 1, 1],
                                      [1, 1, 1, 0, 1, 1], [1, 1, 1, 1, 0, 1], [1, 1, 1, 1, 1, 0]])
BARBELL5_ADJ_MAT = scipy.sparse.csc_matrix([[0, 1, 1, 1, 1, 0, 0, 0, 0, 0],
                                            [1, 0, 1, 1, 1, 0, 0, 0, 0, 0],
                                            [1, 1, 0, 1, 1, 0, 0, 0, 0, 0],
                                            [1, 1, 1, 0, 1, 0, 0, 0, 0, 0],
                                            [1, 1, 1, 1, 0, 1, 0, 0, 0, 0],
                                            [0, 0, 0, 0, 1, 0, 1, 1, 1, 1],
                                            [0, 0, 0, 0, 0, 1, 0, 1, 1, 1],
                                            [0, 0, 0, 0, 0, 1, 1, 0, 1, 1],
                                            [0, 0, 0, 0, 0, 1, 1, 1, 0, 1],
                                            [0, 0, 0, 0, 0, 1, 1, 1, 1, 0],
                                            ])


def test_graph_constructor():
    # Start by constructing the cycle graph on 4 vertices.
    graph = stag.graph.Graph(C4_ADJ_MAT)

    # The graph has 4 vertices and 4 edges
    assert graph.number_of_vertices() == 4
    assert graph.number_of_edges() == 4

    # Check the vertex degrees
    for i in range(4):
        assert graph.degree(i) == 2

    # Now, try creating the complete graph on 6 vertices.
    graph = stag.graph.Graph(K6_ADJ_MAT)
    assert graph.number_of_vertices() == 6
    assert graph.number_of_edges() == 15
    for i in range(4):
        assert graph.degree(i) == 5

    # Now, try the barbell graph
    graph = stag.graph.Graph(BARBELL5_ADJ_MAT)
    assert graph.number_of_vertices() == 10
    assert graph.number_of_edges() == 21
    assert graph.degree(2) == 4
    assert graph.degree(4) == 5


def test_complete_graph():
    # Create a complete graph
    n = 4
    graph = stag.graph.complete_graph(n)
    expected_adjacency_matrix = sp.sparse.csc_matrix([[0, 1, 1, 1],
                                                      [1, 0, 1, 1],
                                                      [1, 1, 0, 1],
                                                      [1, 1, 1, 0]])

    assert graph.number_of_vertices() == 4
    adj_mat_diff = (graph.adjacency() - expected_adjacency_matrix)
    adj_mat_diff.eliminate_zeros()
    assert adj_mat_diff.nnz == 0

    expected_norm_lap = sp.sparse.csc_matrix([[1, -1/3, -1/3, -1/3],
                                              [-1/3, 1, -1/3, -1/3],
                                              [-1/3, -1/3, 1, -1/3],
                                              [-1/3, -1/3, -1/3, 1]])
    norm_lap_diff = (graph.normalised_laplacian() - expected_norm_lap)
    assert(np.all(norm_lap_diff.todense() == pytest.approx(0)))


def test_cycle_graph():
    # Create a cycle graph
    n = 5
    graph = stag.graph.cycle_graph(n)
    expected_adjacency_matrix = sp.sparse.csc_matrix([[0, 1, 0, 0, 1],
                                                      [1, 0, 1, 0, 0],
                                                      [0, 1, 0, 1, 0],
                                                      [0, 0, 1, 0, 1],
                                                      [1, 0, 0, 1, 0]])

    assert graph.number_of_vertices() == 5
    adj_mat_diff = (graph.adjacency() - expected_adjacency_matrix)
    adj_mat_diff.eliminate_zeros()
    assert adj_mat_diff.nnz == 0

    expected_laplacian_matrix = sp.sparse.csc_matrix([[2, -1, 0, 0, -1],
                                                      [-1, 2, -1, 0, 0],
                                                      [0, -1, 2, -1, 0],
                                                      [0, 0, -1, 2, -1],
                                                      [-1, 0, 0, -1, 2]])
    lap_diff = (graph.laplacian() - expected_laplacian_matrix)
    lap_diff.eliminate_zeros()
    assert lap_diff.nnz == 0


def test_adjacency_matrix():
    graph = stag.graph.Graph(BARBELL5_ADJ_MAT)
    adj_mat_diff = (graph.adjacency() - BARBELL5_ADJ_MAT)
    adj_mat_diff.eliminate_zeros()
    assert adj_mat_diff.nnz == 0


def test_symmetry():
    # Generate a large graph from the stochastic block model
    big_graph = stag.random.sbm(1000, 5, 0.8, 0.2)

    # Check that all of the graph matrices are truly symmetric
    assert np.allclose(big_graph.adjacency().toarray(), big_graph.adjacency().toarray().T)

    lap_mat = big_graph.normalised_laplacian()
    lap_mat_dense = lap_mat.toarray()
    assert np.allclose(lap_mat_dense, lap_mat_dense.T)

    lap_mat = big_graph.laplacian()
    lap_mat_dense = lap_mat.toarray()
    assert np.allclose(lap_mat_dense, lap_mat_dense.T)


def test_num_edges():
    # Generate a known graph
    graph = stag.graph.Graph(BARBELL5_ADJ_MAT)

    # Check the number of edges in the graph
    assert graph.number_of_vertices() == 10
    assert graph.number_of_edges() == 21
    assert graph.total_volume() == 42

    # Now create a weighted graph and check the number of edges method.
    adjacency_matrix = scipy.sparse.csc_matrix([[0, 2, 0, 1],
                                                [2, 0, 3, 0],
                                                [0, 3, 0, 1],
                                                [1, 0, 1, 0]])
    graph = stag.graph.Graph(adjacency_matrix)
    assert graph.number_of_vertices() == 4
    assert graph.number_of_edges() == 4
    assert graph.total_volume() == 14


def test_float_weights():
    # Create a graph with floating-point edge weights.
    adjacency_matrix = scipy.sparse.csc_matrix([[0, 2.2, 0, 1],
                                                [2.2, 0, 3.1, 0],
                                                [0, 3.1, 0, 1.09],
                                                [1, 0, 1.09, 0]])
    graph = stag.graph.Graph(adjacency_matrix)
    assert graph.number_of_vertices() == 4
    assert graph.number_of_edges() == 4
    assert graph.total_volume() == pytest.approx(14.78)
    assert graph.degree(0) == pytest.approx(3.2)
    assert graph.degree(2) == pytest.approx(4.19)

    # Check the unweighted degrees
    assert graph.degree_unweighted(0) == 2
    assert graph.degree_unweighted(2) == 2


def test_networkx():
    # Test the methods for converting from and to networkx graphs.
    # Start by constructing a networkx graph
    netx_graph = networkx.generators.barbell_graph(4, 1)
    graph = stag.graph.from_networkx(netx_graph)

    assert graph.number_of_vertices() == 9
    assert graph.number_of_edges() == 14

    expected_adjacency_matrix = sp.sparse.csc_matrix([[0, 1, 1, 1, 0, 0, 0, 0, 0],
                                                      [1, 0, 1, 1, 0, 0, 0, 0, 0],
                                                      [1, 1, 0, 1, 0, 0, 0, 0, 0],
                                                      [1, 1, 1, 0, 0, 0, 0, 0, 1],
                                                      [0, 0, 0, 0, 0, 1, 1, 1, 1],
                                                      [0, 0, 0, 0, 1, 0, 1, 1, 0],
                                                      [0, 0, 0, 0, 1, 1, 0, 1, 0],
                                                      [0, 0, 0, 0, 1, 1, 1, 0, 0],
                                                      [0, 0, 0, 1, 1, 0, 0, 0, 0]])
    adj_mat_diff = (graph.adjacency() - expected_adjacency_matrix)
    adj_mat_diff.eliminate_zeros()
    assert adj_mat_diff.nnz == 0

    # Now, construct a graph using the sgtl Graph object, and convert it to networkx
    graph = stag.graph.Graph(expected_adjacency_matrix)
    netx_graph = graph.to_networkx()

    # Check that the networkx graph looks correct
    assert netx_graph.number_of_nodes() == 9
    assert netx_graph.number_of_edges() == 14
    assert netx_graph.has_edge(0, 1)
    assert netx_graph.has_edge(3, 8)
    assert netx_graph.has_edge(8, 4)
    assert not netx_graph.has_edge(2, 8)


def test_degree_matrix():
    # Construct a graph and get its degree matrix
    g = stag.graph.barbell_graph(4)
    expected_degree_mat = sp.sparse.diags([3, 3, 3, 4, 4, 3, 3, 3])
    deg_mat_diff = g.degree_matrix() - expected_degree_mat
    assert(np.all(deg_mat_diff.todense() == pytest.approx(0)))


def test_inverse_degree_matrix():
    # Construct a graph and get its inverse degree matrix
    g = stag.graph.barbell_graph(4)
    expected_inv_degree_mat = sp.sparse.diags([1/3, 1/3, 1/3, 1/4, 1/4, 1/3, 1/3, 1/3])
    inv_deg_mat_diff = g.inverse_degree_matrix() - expected_inv_degree_mat
    assert(np.all(inv_deg_mat_diff.todense() == pytest.approx(0)))


def test_lazy_random_walk_matrix():
    # Construct a graph
    g = stag.graph.barbell_graph(3)
    expected_rw_mat = sp.sparse.csc_matrix([[1/2, 1/4, 1/6,   0,   0,   0],
                                            [1/4, 1/2, 1/6,   0,   0,   0],
                                            [1/4, 1/4, 1/2, 1/6,   0,   0],
                                            [  0,   0, 1/6, 1/2, 1/4, 1/4],
                                            [  0,   0,   0, 1/6, 1/2, 1/4],
                                            [  0,   0,   0, 1/6, 1/4, 1/2]])
    rw_mat_diff = g.lazy_random_walk_matrix() - expected_rw_mat
    assert(np.all(rw_mat_diff.todense() == pytest.approx(0)))


def test_edge():
    # Test the edge object
    e = stag.graph.Edge(1, 2, 0.3)
    assert(e.v1 == 1)
    assert(e.v2 == 2)
    assert(e.weight == 0.3)

    # Define two edges and make sure they are equal
    e2 = stag.graph.Edge(1, 2, 0.3)
    assert(e == e2)

    # Define a different edge
    e3 = stag.graph.Edge(1, 3, 0.2)
    assert (e2 != e3)


def test_graph_equality():
    g1 = stag.graph.complete_graph(6)
    g2 = stag.graph.complete_graph(6)
    assert g1 == g2

    g3 = stag.graph.Graph(K6_ADJ_MAT)
    assert g2 == g3

    g4 = stag.graph.Graph(BARBELL5_ADJ_MAT)
    assert g2 != g4

    g5 = stag.graph.barbell_graph(5)
    assert g4 == g5
